# -*- coding: utf-8 -*-
import base64
import hashlib
import json
import logging
import logging.handlers
import os
import shutil
import re
import sys
import urllib2
from ConfigParser import SafeConfigParser
from datetime import date

import fcntl
import zlib
from suds.client import Client
from suds.transport.http import HttpAuthenticated

LOCK = None


# Initializes the logger, log, and gives it a name, name
def init_logger(name):
    form = "[%(asctime)s] - %(levelname)8s - %(funcName)15s()] %(message)s"
    datefmt = '%Y-%m-%d %H:%M:%S'
    l = logging.getLogger(name)
    handler = logging.handlers.SysLogHandler(
        facility=logging.handlers.SysLogHandler.LOG_LOCAL0, address='/dev/log')
    l.addHandler(handler)
    formatter = logging.Formatter(form, datefmt=datefmt)
    handler.setFormatter(formatter)


# Authenticates and returns the connection
def connect(usern, passwd):
    trans = HttpAuthenticated(username=usern, password=passwd)
    url = "https://vidyo-replay1.uio.no/replay/services/VidyoReplayContentManagementService?wsdl"
    loc = "https://vidyo-replay1.uio.no/replay/services/VidyoReplayContentManagementService"
    return Client(url, location=loc, transport=trans, faults=False)


# Returns the records from VidyoReplay
# def get_records(client):
def get_records(client):
    try:
        records = client.service.RecordsSearch(sortby='date')
    except Exception as detail:
        s = "Exception: {}".format(detail)
        log.error(s)
        sys.exit(1)
    return records


# Checks if the script has the right amount of arguments
def check_cmd_args(argv):
    if len(argv) is not 2:
        log.error("1 argument needed, found %d", len(argv)-1)
        sys.exit(1)


# Check to see if the recording is owned by the "tsd-import" user.
def check_username(rec):
    username = 'tsd-import'

    if rec.userName == username:
        return True
    else:
        return False


# Returns True if there is not enough space for the record.
def check_usage(path, filesize):
    st = os.statvfs(path)
    free = st.f_bavail * st.f_frsize

    if filesize < free:
        return True
    else:
        return False


# Checks if the script is already running
def single_instance(path):
    global LOCK

    try:
        LOCK = open(path, 'w+')
    except IOError as e:
        s = "IOError: {}".format(e)
        log.error(s)
        sys.exit(1)

    try:
        fcntl.lockf(LOCK, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        s = "the script is already running"
        log.warning(s)
        sys.exit(1)


# Returns true if the user har permissions to read directory d
def read_dir(d):
    if not os.path.isdir(d):
        s = "{} is not a directory".format(d)
        log.error(s)
        sys.exit(1)

    if not os.access(d, os.R_OK):
        s = "No permission to read in directory {}".format(d)
        log.error(s)
        sys.exit(1)

    return True


# Returns true if the user har permissions to write to directory d
def write_dir(d):
    read_dir(d)

    if not os.access(d, os.W_OK):
        s = "No permission to write in directory: {}".format(d)
        log.error(s)
        sys.exit(1)

    return True


# Returns true if the user har permissions to write to file f
def write_file(f):
    if not os.access(f, os.W_OK):
        s = "No permission to write to file: {}".format(f)
        log.error(s)
        sys.exit(1)

    return True


# Parses the config
def read_config(argv):
    config_f = argv[0]

    if not os.path.isfile(config_f):
        s = "{} does not exist".format(config_f)
        log.error(s)
        sys.exit(1)

    parser = SafeConfigParser()
    parser.read(config_f)

    try:
        config = {
            'usern': parser.get('auth', 'username'),
            'passwd': parser.get('auth', 'password'),
            'lock': parser.get('config', 'lock'),
            'dest': parser.get('config', 'videoDest'),
            'delete': parser.getboolean('config', 'delete'),
            'dryrun': parser.getboolean('config', 'dryrun')
            }
    except:
        s = "configuration error: {}".format(sys.exc_info()[0])
        log.error(s)
        sys.exit(1)

    if config['dryrun']:
        print "--- Config ---"
        print config
        print "\n"

    try:
        config['hash'] = parser.get('config', 'hash')   # not mandatory
    except:
        config['hash'] = 'sha256'
        pass

    if not os.path.isfile(config['lock']):
        s = "{} does not exist".format(config['lock'])
        log.error(s)
        sys.exit(1)

    write_file(config['lock'])
    write_dir(config['dest'])

    if not os.path.isdir(config['dest']+"completed"):
        os.mkdir(config['dest']+"completed", 0770)

    return config


# Makes a JSON-file from record metadata
def makeJSON(record, dest, display_name, checksum, hashfunc, dryrun):

    d = {
        'id': record.id,
        'guid': record.guid,
        'tenantName': record.tenantName,
        'userName': record.userName,
        'userFullName': record.userFullName,
        'recordedBy': display_name,
        'dateCreated': record.dateCreated.strftime("%Y-%m-%d %H:%M:%S"),
        'dateCreatedString': record.dateCreatedString,
        # 'endTime': record.endTime,
        'duration': record.duration,
        'resolution': record.resolution,
        'framerate': record.framerate,
        'pin': record.pin,
        'recordScope': record.recordScope,
        'title': record.title,
        'roomName': record.roomName,
        'fileLink': record.fileLink,
        'recorderId': record.recorderId,
        'webcast': record.webcast,
        'tags': record.tags,
        'comments': record.comments,
        'locked': record.locked,
        'externalPlaybackLink': record.externalPlaybackLink,
        'fileSize': record.fileSize,
        'checksum': checksum,
        'hashfunc': hashfunc,
        }

    if dryrun:
        print "--- JSON ---"
        print d
        print "Checksum is always -2 with dryrun"
        print "------------"
        return -1

    if os.path.isfile(dest):
        oldfp = open(dest, 'r')
        oldjson = json.load(oldfp)
        oldfp.close()
    else:
        oldjson = ''

    if oldjson != d:
        with open(dest, 'w') as f:
            json.dump(d, f)


def checksum_and_download(path, hashfunc, f):
    def crc32(path):
        BLOCKSIZE = 65536
        csum = 0
        with open(path, 'ab') as afile:
            buf = f.read(BLOCKSIZE)
            while len(buf) > 0:
                afile.write(buf)
                csum = zlib.crc32(buf, csum)
                buf = f.read(BLOCKSIZE)
        return csum & 0xffffffff

    def sha256(path):
        BLOCKSIZE = 65536
        hasher = hashlib.sha256()
        with open(path, 'ab') as afile:
            buf = f.read(BLOCKSIZE)
            while len(buf) > 0:
                afile.write(buf)
                hasher.update(buf)
                buf = f.read(BLOCKSIZE)
        return hasher.hexdigest()

    if hashfunc == 'crc32':
        return crc32(path)
    else:
        return sha256(path)


def getVideo(config, url):
    username = config['usern']
    passwd = config['passwd']
    dest = config['dest']
    hashfunc = config['hash']
    dryrun = config['dryrun']

    req = urllib2.Request(url)
    base64string = base64.encodestring('%s:%s' % (username, passwd))[:-1]
    req.add_header("Authorization", "Basic %s" % base64string)

    try:
        f = urllib2.urlopen(req)
        condis = f.info()['Content-Disposition']
        result = re.search('\"(.*)\"', condis)
        filename = result.group(1)
    except IOError, e:
        log.error(e)

    if dryrun:
        return filename, -2
    elif os.path.isfile(dest+"completed/"+filename):
        return filename, -1

    checks = checksum_and_download(dest+filename, hashfunc, f)
    return filename, checks


def run(config, client, records):
    username = config['usern']
    passwd = config['passwd']
    dest = config['dest']
    delete = config['delete']
    hashfunc = config['hash']
    dryrun = config['dryrun']

    for record in records[1].records:
        if check_username(record):
            if not check_usage(config['dest'], int(record.fileSize)):
                s = "Disk space full. Record.id: {} Filesize: {}".format(record.id, record.fileSize)
                log.error(s)
                sys.exit(1)

            if dryrun:
                print "--- ID: {} ---".format(record.id)

            filename, checksum = getVideo(config, record.fileLink)
            result = re.search('(.*)_\d\d\d\d', filename)

            if checksum != -1:
                if checksum != -2:
                    shutil.move(dest+filename, dest+"completed/"+filename)
                makeJSON(record, dest +"completed/"+filename[:-4] + '.json',
                    result.group(1), checksum, hashfunc, dryrun)


            # elif checksum == -1 and delete == True:
            #     s = "The file {} already exist. delete = {}".format(dest+filename, delete)
            #     log.warning(s)

            if dryrun and delete:
                print "Record would be deleted"
            elif delete:
                retval = client.service.DeleteRecord(record.id)
                if retval[1] != "OK":
                    s = "DeleteRecord({}) should return (200, OK). Returned {}.".format(
                        record.id, retval)
                    log.error(s)
            if dryrun:
                print "\n"


if __name__ == "__main__":
    init_logger('log')
    log = logging.getLogger('log')
    check_cmd_args(sys.argv)
    config = read_config(sys.argv[1:])
    single_instance(config['lock'])

    client = connect(config['usern'], config['passwd'])
    records = get_records(client)
    run(config, client, records)
